html
{
  -webkit-font-smoothing: antialiased;
}

body
{
  font: 14px/24px 'Open Sans', sans-serif;
}

#container
{
	margin: 0 30px;
	background: #fff;
}

#header
{
	background: #fff;
	padding-left: 10px;
}

#header h1 { margin: 0; }

#navigation
{
	float: left;
	width: 100%;
	background: #fff;
	font-family: 'Droid Sans', sans-serif;
	font-size: 15px;
}

#navigation ul
{
	margin: 0;
	padding: 0;
}

#navigation ul li
{
	list-style-type: none;
	display: inline;
}

#navigation li a
{
	display: block;
	float: left;
	padding: 5px 10px;
	color: #666666;
	text-decoration: none;
	border-right: 1px solid #fff;
}

#navigation li a.active { color: #79AE3D; font-weight: bold; }

#navigation li a:hover { color: #79AE3D; text-decoration: underline; }

#content
{
	clear: left;
	padding: 10px;
}

#content h2
{
	color: #000;
	font-size: 160%;
	margin: 0 0 .5em;
}

#footer
{
	background: #f;
	text-align: right;
	padding: 20px;
	height: 1%;
}

.logo-title-a {
  text-decoration: none;
}
.logo-title {
  font-size: 20px;
}
.logo-title span {
  color:#79AE3D;
  font-size:27px;
}

.mzn-arg {
  font-family: monospace;
  font-weight: bold;
  font-size: 110%;
}
.mzn-parm {
  font-style: italic;
}

.mzn-group-level-0 .mzn-group-name {
  font-size: 150%; font-weight: bold;  
}
.mzn-group-level-1 .mzn-group-name {
  font-size: 120%; font-weight: bold;  
}

.mzn-decl-type-heading {
  display: none;
/*  font-size: 150%; font-weight: bold;*/
}

.mzn-fundecl, .mzn-vardecl {
  width: 75%;
  margin-bottom: 10px;
  margin-left: 5%;
  border: 1px solid #C0C0C0;
}

.mzn-fundecl-code, .mzn-vardecl-code {
  font-family: monospace;
/*  font-weight: bold;*/
  font-size: 110%;
  white-space: pre;
  margin-bottom: 2px;
  background-color: #EAEAEA;
  border-bottom: 1px solid #C0C0C0;
  padding: 3px;
}

.mzn-fundecl-more {
  float: right;
  text-decoration: none;
  color: black;
}

.mzn-fundecl-more-code {
  display: none;
}

.mzn-fundecl-equals {
  display: none;
}

.mzn-fundecl-reveal-equals {
  display: inline;
}

.mzn-fundecl-code .mzn-fundecl-reveal-code {
  font-size: 10pt;
}

.mzn-fundecl-body {
  padding-left: 1em;
  font-size: 110%;
}

.mzn-fundecl-reveal-code {
 /* font-family: sans-serif;
     font-size: 11pt;
    */
  display: block;
  margin-top: 5px;
}

.mzn-fundecl-doc, .mzn-vardecl-doc {
  padding-left: 3px;
}

.mzn-id {
  font-weight: bold;
}

.mzn-fn-id {
	font-style: italic;
	color: blue;
}

a.mzn-nav-up, a.mzn-nav-next, a.mzn-nav-prev {
  text-decoration: none;
}